package gui.report.image;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.awt.image.BufferedImageOp;
import java.awt.image.ByteLookupTable;
import java.awt.image.ColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.LookupOp;
import java.awt.image.Raster;
import java.awt.image.SampleModel;
import java.awt.image.WritableRaster;
import java.util.Iterator;
import java.util.Vector;

import javax.media.jai.PlanarImage;
import javax.media.jai.RasterFactory;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.SwingConstants;

/**
 * An output widget for a PlanarImage. ImageDisplay subclasses
 * javax.swing.JComponent, and can be used in any context that calls for a
 * JComponent. It monitors resize and update events and automatically requests
 * tiles from its source on demand.
 * <p>
 * Due to the limitations of BufferedImage, only TYPE_BYTE of band 1, 2, 3, 4,
 * and TYPE_USHORT of band 1, 2, 3 images can be displayed using this widget.
 * 
 * @author Daniel Rice
 * @author Dennis Sigel
 */

public class ImageDisplay extends JComponent implements MouseListener,
		MouseMotionListener {

	/**
	 * 
	 */
	private static final long serialVersionUID = -2679467019690145221L;

	/**
	 * The source PlanarImage.
	 * 
	 * @uml.property name="source"
	 * @uml.associationEnd multiplicity="(0 1)"
	 */
	protected PlanarImage source;

	/** The image's SampleModel. */
	protected SampleModel sampleModel;

	/** The image's ColorModel or one we supply. */
	protected ColorModel colorModel;

	/** The image's min X tile. */
	protected int minTileX;

	/** The image's max X tile. */
	protected int maxTileX;

	/** The image's min Y tile. */
	protected int minTileY;

	/** The image's max Y tile. */
	protected int maxTileY;

	/** The image's tile width. */
	protected int tileWidth;

	/** The image's tile height. */
	protected int tileHeight;

	/** The image's tile grid X offset. */
	protected int tileGridXOffset;

	/** The image's tile grid Y offset. */
	protected int tileGridYOffset;

	protected int originX = 0;

	protected int originY = 0;

	protected int shift_x = 0;

	protected int shift_y = 0;

	/**
	 * @uml.property name="odometer"
	 * @uml.associationEnd multiplicity="(0 1)"
	 */
	protected JLabel odometer = null;

	protected int componentWidth;

	protected int componentHeight;

	/** Brightness control */
	protected BufferedImageOp biop = null;

	protected boolean brightnessEnabled = false;

	protected int brightness = 0;

	protected byte[] lutData;

	/**
	 * @uml.property name="annotationVector"
	 * @uml.associationEnd elementType="java.awt.Rectangle" multiplicity="(0
	 *                     -1)"
	 */
	private Vector<Rectangle> annotationVector = new Vector<Rectangle>();

	

	private Point startPoint = new Point(0, 0);

	private Point endPoint;

	/**
	 * @uml.property name="annotationRec" multiplicity="(0 1)"
	 */
	private Rectangle annotationRec = new Rectangle();

	/**
	 * @uml.property name="mouseListener"
	 * @uml.associationEnd inverse="adaptee:swpatho.ap1.editor.This_mouseAdapter"
	 *                     multiplicity= "(1 1)"
	 */

	/** Initializes the ImageDisplay. */
	private synchronized void initialize() {
		if (source == null)
			return;

		componentWidth = source.getWidth();
		componentHeight = source.getHeight();

		setPreferredSize(new Dimension(componentWidth, componentHeight));

		this.sampleModel = source.getSampleModel();

		// First check whether the opimage has already set a suitable ColorModel
		this.colorModel = source.getColorModel();
		if (this.colorModel == null) {
			// If not, then create one.
			this.colorModel = PlanarImage.createColorModel(this.sampleModel);
			if (this.colorModel == null) {
				throw new IllegalArgumentException("no color model");
			}
		}

		minTileX = source.getMinTileX();
		maxTileX = source.getMinTileX() + source.getNumXTiles() - 1;
		minTileY = source.getMinTileY();
		maxTileY = source.getMinTileY() + source.getNumYTiles() - 1;
		tileWidth = source.getTileWidth();
		tileHeight = source.getTileHeight();
		tileGridXOffset = source.getTileGridXOffset();
		tileGridYOffset = source.getTileGridYOffset();

		this.addMouseMotionListener(new MouseMotionListener() {
			public void mouseDragged(MouseEvent arg0) {
				ImageDisplay.this.mouseDragged(arg0);
			}

			public void mouseMoved(MouseEvent arg0) {
				ImageDisplay.this.mouseMoved(arg0);
			}
		});
		this.addMouseListener(new MouseListener() {
			public void mousePressed(MouseEvent arg0) {
				ImageDisplay.this.mousePressed(arg0);
			}

			public void mouseReleased(MouseEvent arg0) {
				ImageDisplay.this.mouseReleased(arg0);

			}

			public void mouseClicked(MouseEvent arg0) {
				// TODO Auto-generated method stub

			}

			public void mouseEntered(MouseEvent arg0) {
			}

			public void mouseExited(MouseEvent arg0) {
				// TODO Auto-generated method stub

			}
		});
	}

	/**
	 * Default constructor
	 */
	/*public ImageDisplay() {
		super();
		source = null;

		lutData = new byte[256];

		for (int i = 0; i < 256; i++) {
			lutData[i] = (byte) i;
		}

		componentWidth = 64;
		componentHeight = 64;
		setPreferredSize(new Dimension(componentWidth, componentHeight));
		setOrigin(0, 0);
	}*/

	/**
	 * Constructs an ImageDisplay to display a PlanarImage.
	 * 
	 * @param im
	 *            PlanarImage
	 */
	public ImageDisplay(PlanarImage im) {
		super();
		source = im;
		initialize();

		lutData = new byte[256];

		for (int i = 0; i < 256; i++) {
			lutData[i] = (byte) i;
		}

		setOrigin(0, 0);
	}

	/**
	 * Constructs an ImageDisplay of fixed size (no image)
	 * 
	 * @param width -
	 *            display width
	 * @param height -
	 *            display height
	 */
	public ImageDisplay(int width, int height) {
		super();
		source = null;

		lutData = new byte[256];

		for (int i = 0; i < 256; i++) {
			lutData[i] = (byte) i;
		}

		componentWidth = width;
		componentHeight = height;
		setPreferredSize(new Dimension(componentWidth, componentHeight));
		setOrigin(0, 0);
	}

	public void addAnnotationGraphic(Rectangle annotationRec) {
		annotationVector.add(annotationRec);
		annotationRec = null;
	}

	

	/**
	 * @uml.property name="annotationRec"
	 */
	public Rectangle getAnnotationRec() {
		return annotationRec;
	}

	/**
	 * @uml.property name="annotationVector"
	 */
	public Vector getAnnotationVector() {
		return annotationVector;
	}

	/**
	 * Changes the source image to a new PlanarImage.
	 * 
	 * @param im
	 *            PlanarImage
	 */
	public void set(PlanarImage im) {
		source = im;
		initialize();
		repaint();
	}

	public void set(PlanarImage im, int x, int y) {
		source = im;
		initialize();
		setOrigin(x, y);
	}

	public PlanarImage getImage() {
		return source;
	}

	/**
	 * @uml.property name="odometer"
	 */
	public final JLabel getOdometer() {
		if (odometer == null) {
			odometer = new JLabel();
			odometer.setVerticalAlignment(SwingConstants.CENTER);
			odometer.setHorizontalAlignment(SwingConstants.LEFT);
			odometer.setText(" ");
			addMouseListener(this);
			addMouseMotionListener(this);
		}

		return odometer;
	}

	/** Provides panning */
	public final void setOrigin(int x, int y) {
		// shift to box origin
		
		
		
		originX = -x;
		originY = -y;
		repaint();
	}

	public int getXOrigin() {
		return originX;
	}

	public int getYOrigin() {
		return originY;
	}

	/** Records a new size. Called by the AWT. */
	@Override
	public void setBounds(int x, int y, int width, int height) {
		Insets insets = getInsets();
		int w;
		int h;

		if (source == null) {
			w = width;
			h = height;
		} else {
			w = source.getWidth();
			h = source.getHeight();

			if (width < w) {
				w = width;
			}

			if (height < h) {
				h = height;
			}
		}

		componentWidth = w + insets.left + insets.right;
		componentHeight = h + insets.top + insets.bottom;

		super.setBounds(x + shift_x, y + shift_y, componentWidth,
				componentHeight);
	}

	@Override
	public void setLocation(int x, int y) {
		shift_x = x;
		shift_y = y;
		super.setLocation(x, y);
	}

	private final int XtoTileX(int x) {
		return (int) Math.floor((double) (x - tileGridXOffset) / tileWidth);
	}

	private final int YtoTileY(int y) {
		return (int) Math.floor((double) (y - tileGridYOffset) / tileHeight);
	}

	private final int TileXtoX(int tx) {
		return tx * tileWidth + tileGridXOffset;
	}

	private final int TileYtoY(int ty) {
		return ty * tileHeight + tileGridYOffset;
	}

	/**
	 * Paint the image onto a Graphics object. The painting is performed
	 * tile-by-tile, and includes a grey region covering the unused portion of
	 * image tiles as well as the general background. At this point the image
	 * must be byte data.
	 */
	@Override
	public synchronized void paintComponent(Graphics g) {

		Graphics2D g2D = null;
		if (g instanceof Graphics2D) {
			g2D = (Graphics2D) g;
		} else {
			return;
		}

		// if source is null, it's just a component
		if (source == null) {
			g2D.setColor(getBackground());
			g2D.fillRect(0, 0, componentWidth, componentHeight);
			return;
		}

		int transX = -originX;
		int transY = -originY;

		// Get the clipping rectangle and translate it into image coordinates.
		Rectangle clipBounds = g.getClipBounds();

		if (clipBounds == null) {
			clipBounds = new Rectangle(0, 0, componentWidth, componentHeight);
		}

		// clear the background (clip it) [minimal optimization here]
		if (transX > 0 || transY > 0
				|| transX < (componentWidth - source.getWidth())
				|| transY < (componentHeight - source.getHeight())) {
			g2D.setColor(getBackground());
			g2D.fillRect(0, 0, componentWidth, componentHeight);
		}

		clipBounds.translate(-transX, -transY);

		// Determine the extent of the clipping region in tile coordinates.
		int txmin, txmax, tymin, tymax;
		int ti, tj;

		txmin = XtoTileX(clipBounds.x);
		txmin = Math.max(txmin, minTileX);
		txmin = Math.min(txmin, maxTileX);

		txmax = XtoTileX(clipBounds.x + clipBounds.width - 1);
		txmax = Math.max(txmax, minTileX);
		txmax = Math.min(txmax, maxTileX);

		tymin = YtoTileY(clipBounds.y);
		tymin = Math.max(tymin, minTileY);
		tymin = Math.min(tymin, maxTileY);

		tymax = YtoTileY(clipBounds.y + clipBounds.height - 1);
		tymax = Math.max(tymax, minTileY);
		tymax = Math.min(tymax, maxTileY);
		Insets insets = getInsets();

		// Loop over tiles within the clipping region
		for (tj = tymin; tj <= tymax; tj++) {
			for (ti = txmin; ti <= txmax; ti++) {
				int tx = TileXtoX(ti);
				int ty = TileYtoY(tj);

				Raster tile = source.getTile(ti, tj);
				if (tile != null) {
					DataBuffer dataBuffer = tile.getDataBuffer();

					WritableRaster wr = Raster.createWritableRaster(
							sampleModel, dataBuffer, null);

					BufferedImage bi = new BufferedImage(colorModel, wr,
							colorModel.isAlphaPremultiplied(), null);

					// correctly handles band offsets
					if (brightnessEnabled == true) {
						SampleModel sm = sampleModel
								.createCompatibleSampleModel(tile.getWidth(),
										tile.getHeight());

						WritableRaster raster = RasterFactory
								.createWritableRaster(sm, null);

						BufferedImage bimg = new BufferedImage(colorModel,
								raster, colorModel.isAlphaPremultiplied(), null);

						// don't move this code
						ByteLookupTable lutTable = new ByteLookupTable(0,
								lutData);
						LookupOp lookup = new LookupOp(lutTable, null);
						lookup.filter(bi, bimg);

						g2D.drawImage(bimg, biop, tx + transX + insets.left, ty
								+ transY + insets.top);
					} else {
						AffineTransform transform;

						transform = AffineTransform.getTranslateInstance(tx
								+ transX + insets.left, ty + transY
								+ insets.top);

						g2D.drawRenderedImage(bi, transform);
					}
				}
			}
		}
		g2D.setColor(Color.red);
		// paint annotations
		for (Rectangle rec : annotationVector) {
			Rectangle rec2 = (Rectangle) rec.clone();

			// image can be moved by panner --> coordinates have to be adjusted
			rec2.x -= getXOrigin();
			rec2.y -= getYOrigin();
	
			g2D.draw(rec2);
		}
		if (annotationRec != null){
			Rectangle rec2 = (Rectangle)annotationRec.clone();
			rec2.x -= getXOrigin();
			rec2.y -= getYOrigin();
	
			g2D.draw(rec2);
		}
	}

	// mouse interface
	public final void mouseEntered(MouseEvent e) {
	}

	public final void mouseExited(MouseEvent e) {
	}

	public void mousePressed(MouseEvent e) {
		System.out.println("mousePressed()");
		/*
		 * if (!annotationConfirmed) if (annotationVector.size() > 0)
		 * annotationVector.remove(annotationVector.size() - 1);
		 */
		
		startPoint = e.getPoint();
		startPoint.x += getXOrigin();
		startPoint.y += getYOrigin();
		System.out.println(startPoint.x + " " + startPoint.y);
		if (odometer != null) {
			String output = " (" + startPoint.x + ", " + startPoint.y + ")";
			odometer.setText(output);
		}
		
		for(Iterator<Rectangle> i = annotationVector.iterator(); i.hasNext();){
			if(i.next().contains(startPoint));

		}
		
	}

	public final void mouseReleased(MouseEvent e) {
		endPoint = e.getPoint();
		endPoint.x += getXOrigin();
		endPoint.y += getYOrigin();
		Dimension d = new Dimension(
				(int) (endPoint.getX() - startPoint.getX()), (int) (endPoint
						.getY() - startPoint.getY()));
		annotationRec = new Rectangle(startPoint, d);
		this.repaint();

		if (odometer != null) {
			String output = " (" + endPoint.x + ", " + endPoint.y + ")";
			odometer.setText(output);
		}
	}

	public final void mouseClicked(MouseEvent e) {
		System.out.println("mouseClicked()");
	}

	public final void mouseMoved(MouseEvent e) {
		Point p = e.getPoint();
		if (e.getModifiers() == MouseEvent.MOUSE_PRESSED)
			System.out.println("Button1 pressed!");
		if (odometer != null) {
			String output = " (" + p.x + ", " + p.y + ")";
			odometer.setText(output);
		}
	}

	public final void mouseDragged(MouseEvent e) {
		endPoint = e.getPoint();
		endPoint.x += getXOrigin();
		endPoint.y += getYOrigin();
		Dimension d = new Dimension(
				(int) (endPoint.getX() - startPoint.getX()), (int) (endPoint
						.getY() - startPoint.getY()));
		annotationRec = new Rectangle(startPoint, d);
		this.repaint();
	}
}
