
package gui.suchen.graph;

import gui.search.ReportInfo;
import io.JenaLib;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;

import services.suchen.ReportSuchErgebnis;
import services.suchen.ReportSuchKonfiguration;
import services.suchen.SuchService;
import tools.ConcurrentCache;
import tools.Enums.Relation;
import tools.Enums.Section;

import com.hp.hpl.jena.ontology.OntClass;

import controll.RelationCacheControll;

public class KonzeptModel
{
	/* | = - = - = - = - = - = - = - = - = - = /-||=||-\ = - = - = - = - = - = - = - = - = - = | */
	/* | = == === ==== ===== ====== ======= == Constants == ======= ====== ===== ==== === == = | */
	/* | = - = - = - = - = - = - = - = - = - = \-||=||-/ = - = - = - = - = - = - = - = - = - = | */

	private final String									url;
	private final FutureTask<OntClass>						ontologyClassFuture;
	private final FutureTask<ReportSuchErgebnis>			reportIDsFutureTask;

	private OntClass ontologyClass;
	
	private final ConcurrentCache<Relation, List<String>>	parents		= new ConcurrentCache<Relation, List<String>>();
	private final ConcurrentCache<Relation, List<String>>	children	= new ConcurrentCache<Relation, List<String>>();
	private LoadListener									listener;

	/* | = - = - = - = - = - = - = - = - = - = /-||=||-\ = - = - = - = - = - = - = - = - = - = | */
	/* | = == === ==== ===== ====== ======== Constructors = ======= ====== ===== ==== === == = | */
	/* | = - = - = - = - = - = - = - = - = - = \-||=||-/ = - = - = - = - = - = - = - = - = - = | */

	
	
	public KonzeptModel(final String uri) {
		// Preconditions:
		assert uri != null;

		// Implementation:
		this.url = uri;

		ontologyClass = JenaLib.getClassFromString(uri);
		
		ontologyClassFuture = new FutureTask<OntClass>(new Callable<OntClass>() {
			public OntClass call() {
				return JenaLib.getClassFromString(uri);
			}
		});

		reportIDsFutureTask = new FutureTask<ReportSuchErgebnis>(
			new Callable<ReportSuchErgebnis>() {
				public ReportSuchErgebnis call() {
					ReportSuchKonfiguration reportSuchKonfiguration = new ReportSuchKonfiguration(
						Arrays.asList(Section.values()), false);
					reportSuchKonfiguration.setConcepts(getURL());
					ReportSuchErgebnis sucheReports = new SuchService().sucheReports(reportSuchKonfiguration);
					announceFinish();
					return sucheReports;
				}
			});

		Thread thread1 = new Thread(ontologyClassFuture);
		Thread thread2 = new Thread(reportIDsFutureTask);

		thread1.setDaemon(true);
		thread2.setDaemon(true);
		thread1.setPriority(Thread.MIN_PRIORITY);
		thread2.setPriority(Thread.MIN_PRIORITY);

		thread1.start();
		thread2.start();
	}

	/* | = - = - = - = - = - = - = - = - = - = /-||=||-\ = - = - = - = - = - = - = - = - = - = | */
	/* | = == === ==== ===== ====== ======= === Methods === ======= ====== ===== ==== === == = | */
	/* | = - = - = - = - = - = - = - = - = - = \-||=||-/ = - = - = - = - = - = - = - = - = - = | */

	protected void announceFinish() {
		if (listener != null)
			listener.onLoaded();
	}

	public String getURL() {
		return url;
	}

	public ReportSuchErgebnis getReportErgebnisse() {
		try {
			return reportIDsFutureTask.get();
		}
		catch (InterruptedException e) {
			throw new RuntimeException(e);
		}
		catch (ExecutionException e) {
			throw new RuntimeException(e);
		}
	}

	public Collection<ReportInfo> getReports() {
		return getReportErgebnisse().getReports();
	}

	public boolean reportsLoaded() {
		return reportIDsFutureTask.isDone();
	}

	private OntClass getOntologyClass() {
		try {
			return ontologyClassFuture.get();
		}
		catch (InterruptedException e) {
			throw new RuntimeException(e);
		}
		catch (ExecutionException e) {
			throw new RuntimeException(e);
		}
	}

	/* | = - = - = - = - = - = - = - = - = - = /-||=||-\ = - = - = - = - = - = - = - = - = - = | */
	/* | = == === ==== ===== ====== ======== Inner Classes ======== ====== ===== ==== === == = | */
	/* | = - = - = - = - = - = - = - = - = - = \-||=||-/ = - = - = - = - = - = - = - = - = - = | */

	
	public List<String> getNeighbors(boolean findParents, Relation relation) {
		// Preconditions:
		assert relation != null;

		// Implementation:
		ConcurrentCache<Relation, List<String>> cache = findParents ? parents : children;

		List<String> result = cache.get(relation);
		if (result != null)
			return result;

		// Daten nicht vorhanden, nachladen
		cache.writeLock(relation);
		try {
			// evtl. zwischenzeitlich geladen:
			result = cache.get(relation);
			if (result == null)
				cache.put(relation, result = 
					(findParents) ? RelationCacheControll.findParents(relation,url):
						RelationCacheControll.findChildren(relation, url)
					
					
				
				);

			return result;
		}
		finally {
			cache.writeUnLock(relation);
		}
	}

	public void setListener(LoadListener listener) {
		this.listener = listener;
	}
}
