package services.similarity;


import gui.search.ReportInfo;

import io.JenaLib;
import io.SQLDBAccess;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.Map;

import similarity.documents.Document;
import similarity.ontology.Concept;
import tools.Enums.Section;

import com.hp.hpl.jena.ontology.Individual;
import com.hp.hpl.jena.ontology.OntResource;
import com.hp.hpl.jena.rdf.model.Property;

/**
 * This class represents a pathologic report.
 * Parts are described by swpatho.Enums.Section.
 * 
 * @author Rafael Grote <grote@inf.fu-berlin.de>
 * @see similarity.documents.Document
 * @see swpatho.Enums.Section
 */
public class Report implements Document<Section>
{
    /**
     * Report ID
     */
    private final String id;
    
    /**
     * Cached individuals to improve performance.
     * Already converted to concepts and mapped into sections.
     * 
     * Do not access directly! Use getConceptMap() instead.
     */
    private EnumMap<Section,HashMap<Concept,Integer>> cachedConceptMap;

    /**
     * Create report with given id.
     * 
     * @param id report id
     */
    public Report(String id)
    {
        this.id = id;
    }
    
    /**
     * Get <b>all</b> concepts contained in this report.
     * 
     * @see similarity.documents.Document#getConcepts()
     */
    public Map<Concept,Integer> getConcepts()
    {
        Map<Concept,Integer> concepts = new HashMap<Concept,Integer>();
        
        for (Map<Concept,Integer> part : getConceptMap().values())
        {
            for (Map.Entry<Concept,Integer> entry : part.entrySet())
            {
                Integer count = concepts.get(entry.getKey());
                
                if (count == null)
                    count = new Integer(0);
                
                concepts.put(entry.getKey(),count+entry.getValue());
            }
        }
        
        return concepts;
    }

    /**
     * Get concepts contained in the specified section of this report.
     * 
     * @see similarity.documents.Document#getConcepts(Enum)
     */
    public Map<Concept,Integer> getConcepts(Section section)
    {
        if (getConceptMap().containsKey(section))
            return getConceptMap().get(section);
        
        else
            return Collections.emptyMap();
    }
    
    @Override
    public String toString()
    {
        return id;
    }
    
    public boolean equals(Report other)
    {
        return id.equals(other.id);
    }
    
    /**
     * Get a map of all concepts mapped to related sections.
     * 
     * @return all concepts mapped to related sections
     */
    protected final EnumMap<Section,HashMap<Concept,Integer>> getConceptMap()
    {
        // if not yet done, create and fill map
        if (cachedConceptMap == null)
        {
            cachedConceptMap = new EnumMap<Section,HashMap<Concept,Integer>>(Section.class);

            // map concepts to sections
            for (Individual individual : getIndividuals())
            {
                // get section which contains this individual
                Section section = sectionOf(individual);
                
                // create section entry, if not yet existing
                if (!cachedConceptMap.containsKey(section))
                    cachedConceptMap.put(section, new HashMap<Concept,Integer>());
                
                // get concept map for section
                HashMap<Concept,Integer> concepts = cachedConceptMap.get(section);
                
                // get concept for type of individual
                Concept type = new Concept(individual.getRDFType(true));
                
                // get old concept count
                Integer count = concepts.get(type);
                
                // no entry?
                if (count == null)
                    count = new Integer(0);
                
                // increment and write back
                concepts.put(type,count+1);
            }
        }

        return cachedConceptMap;
    }
    
    
    /*
     * DB access
     */
    
    /**
     * Load individuals for this report from db.
     */
    public Collection<Individual> getIndividuals()
    {
        try
        {
            ResultSet rs = SQLDBAccess.query(
                		
                		
                		"SELECT individual FROM Diagnosen WHERE befund="+id+";"
                		
                
                );
            
            Collection<Individual> individuals = new ArrayList<Individual>();
            
            while (rs.next())
            {
       
            	
                Individual next =
                	JenaLib.getIndividualFromString(rs.getString("individual"));
                	
                	
             
                assert next != null;
                
                if (next != null)
                    individuals.add(next);
            }
            
            return individuals;
        }
        catch (SQLException e)
        {
            return Collections.emptyList();
        }
    }
    
    /**
     * Create ReportInfo and fill with data from db.
     * 
     * @return related ReportInfo object
     */
    public ReportInfo getReportInfo()
    {
        ReportInfo info = new ReportInfo(id);
        
        try
        {
            ResultSet rs = SQLDBAccess.
                query("SELECT age, gender FROM Befunde WHERE befund="+id+";");
            
            if (rs.next())
            {
                info.setAge(rs.getInt("age"));
                info.setGender(rs.getString("gender"));
            }
        }
        catch (SQLException e) {/*ignore*/}
        
        return info;
    }
    
    
    /*
     * Tools
     */

    /**
     * Get the section containing the given individual.
     * 
     * @param individual a individual
     * @return section which contains the individual
     */
    public static Section sectionOf(Individual individual)
    {
        Property containedBy = JenaLib.getUMLSContainedByProperty();
        
        return Section.getSection(
                ((OntResource) individual.getPropertyValue(containedBy).as(OntResource.class))
                .getRDFType(true).getLocalName());
    }
         
}
