/**
 * Services for searching similar documents.
 */
package services.similarity;


import gui.search.ReportInfo;

import java.util.Collection;
import java.util.Collections;
import java.util.EnumMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import services.coordination.Parameter;
import services.coordination.ServiceID;
import services.coordination.ServiceProvider;
import services.suchen.FortschrittInfo;
import services.suchen.ReportSuchErgebnis;
import services.suchen.ReportSuchKonfiguration;
import similarity.concepts.Engine;
import similarity.concepts.matcher.util.CachedMatcher;
import similarity.documents.Matcher;
import similarity.documents.PairingMatcher;
import similarity.documents.Result;
import tools.Enums.Section;
import controll.MainControll;

/**
 * @author Rafael Grote <grote@inf.fu-berlin.de>
 *
 */
public class SimilarReportsSearch implements ServiceProvider
{
    /**
     * Minimum similarity value for a report to be considered similar.
     */
    public static final float MIN_SIMILARITY = 0.50f;
    
    /**
     * Should the cache be autimatically filled?
     * If <code>true</code> missing values will be stored in cache on access.
     */
    public static final boolean FILL_CACHE = true;
    
    /**
     * Weighting for the sections.
     */
    public static final Map<Section,Float> WEIGHTING = new EnumMap<Section,Float>(Section.class);
    
    static
    {
        WEIGHTING.put(Section.DIAGNOSE,5f);
        WEIGHTING.put(Section.MAKROSKOPIE,3f);
        WEIGHTING.put(Section.MIKROSKOPIE,3f);
        WEIGHTING.put(Section.KOMMENTAR,1f);
    }
    
    /**
     * Document similarity matcher.
     */
    public static final Matcher<Section> DOC_MATCHER = 
        new PairingMatcher<Section>(new Engine(new CachedMatcher(new SqlCache(),Engine.DEFAULT_MATCHER,FILL_CACHE)),WEIGHTING);
    

    /**
     * @param id must be ServiceID.PERFORM_SIMILARITY_SEARCH
     * @param param report id (must have type String)
     * @see services.coordination.ServiceProvider#performService(services.coordination.ServiceID, java.lang.Object)
     */
    public void performService(ServiceID id, final Parameter param)
    {
        assert id == ServiceID.PERFORM_SIMILARITY_SEARCH;

        
        // TODO improve thread behaviour, fix servicecoordinator (see ServiceCoordinator)
        new Thread(new Runnable()
        {
            public void run()
            {
                MainControll.SERVICECOORDINATOR.requestService(ServiceID.SHOW_PROGRESS,
                        
                		new Parameter(null,
                		new FortschrittInfo(true, "Suche")));
                
                MainControll.SERVICECOORDINATOR.requestService(ServiceID.SHOW_SIMILAR_REPORTS,
new Parameter (param.source, findSimilarReports(
		
		(String)param.value
    	

                )));
                
                MainControll.SERVICECOORDINATOR.requestService(ServiceID.SHOW_PROGRESS,
                		new Parameter(null,
                		
                        new FortschrittInfo(false)));
            }
        })
        .start();
    }
    
    /**
     * Find reports which are similar to the report with specified id.
     */
    public ReportSuchErgebnis findSimilarReports(String reportID)
    {
        return findSimilarReports(ReportFactory.get(reportID));
    }
    
    /**
     * Find reports which are similar to the specified.
     */
    public ReportSuchErgebnis findSimilarReports(Report report)
    {
        Set<ReportInfo> similars = new HashSet<ReportInfo>();

        for (Report r : ReportFactory.getAll())
        {  
        
        	
            Result<Section> sim = DOC_MATCHER.similarity(report,r);
            assert !report.equals(r) || sim.getTotal() == 1;
            
            if (sim.getTotal()>=MIN_SIMILARITY && !report.equals(r))
            {
            	
            	
                ReportInfo info = r.getReportInfo();
                info.setSimResult(sim);                
                similars.add(info);
            }
        }
        
        return new SimilarReportsSearchResult(report.getReportInfo(),similars);
    }
    
    /**
     * 
     * @author Rafael Grote <grote@inf.fu-berlin.de>
     *
     */
    public static class SimilarReportsSearchResult implements ReportSuchErgebnis
    {
        /**
         * The report for which similars were searched.
         */
        private final ReportInfo report;
        
        /**
         * Similar reports.
         */
        private final Set<ReportInfo> similars;
        
        /**
         * Simple configuration which sorts by similarity
         */
        private final ReportSuchKonfiguration conf;
        
        /**
         * @param report report for which similars were searched
         * @param similars similar reports
         */
        public SimilarReportsSearchResult(ReportInfo report, Set<ReportInfo> similars)
        {
            this.report = report;
            this.similars = similars;
            
            conf = new ReportSuchKonfiguration(getSections(),false);
            conf.setSortiermodus(ReportSuchKonfiguration.Sortiermodus.NACH_AEHNLICHKEIT);
            conf.setConcepts();
        }

        /**
         * @return collection with a single entry
         * @see services.suchen.ReportSuchErgebnis#getSections()
         */
        public Collection<Section> getSections()
        {
            return Collections.singleton(Section.KOMMENTAR);
        }

        /**
         * @return configuration
         * @see services.suchen.ReportSuchErgebnis#getKonfiguration()
         */
        public ReportSuchKonfiguration getKonfiguration()
        {
            return conf;
        }

        /**
         * Same as getReports().
         * The parameter is irrelevant!
         *  
         * @see services.suchen.ReportSuchErgebnis#getReports(swpatho.Enums.Section)
         */
        public Set<ReportInfo> getReports(Section _)
        {
            return getReports();
        }

        /**
         * Get the found similar reports.
         * 
         * @return similar reports
         * @see services.suchen.ReportSuchErgebnis#getReports()
         */
        public Set<ReportInfo> getReports()
        {
            return similars; 
        }
        
        /**
         * Get the report for which similars were searched.
         */
        public ReportInfo getOriginalReport()
        {
            return report;
        }
    }
}
