package services.similarity;

import io.SQLDBAccess;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.logging.Logger;

import similarity.concepts.matcher.util.Cache;
import similarity.ontology.Concept;

/**
 * This is a connector to a database where similarity values are stored.
 * 
 * @author Rafael Grote <grote@inf.fu-berlin.de>
 */
public class SqlCache implements Cache {
	/**
	 * Name of the table used for the cache.
	 */
	public static final String TABLE = "Similarity";

	/**
	 * Maximum length for a concept uri in database.
	 */
	public static final int MAX_URI_LENGTH = 128;

	/**
	 * Read similarity value from database for the given concepts. c1 and c2 can
	 * be switched. <br>
	 * <b>Note: If there is an error on database access, this method will return
	 * <code>-1</code>.</b>
	 * 
	 * @param c1
	 *            a concept
	 * @param c2
	 *            another concept
	 * @return similarity value, <code>-1</code> if no such entry,
	 *         <code>-1</code> if an database error occures
	 * @see similarity.concepts.matcher.util.Cache#get(similarity.ontology.Concept,
	 *      similarity.ontology.Concept)
	 */
	public float get(Concept c1, Concept c2) {
		try {
			ResultSet rs = SQLDBAccess.query("SELECT sim FROM " + TABLE
					+ " WHERE " + "c1='" + c1.getURI() + "' AND c2='"
					+ c2.getURI() + "' OR " + "c1='" + c2.getURI()
					+ "' AND c2='" + c1.getURI() + "';"

			);

			if (!rs.next())
				return -1;

			return rs.getFloat("sim");
		} catch (SQLException e) {
			Logger.global.fine("Database read access failed (" + e.getMessage()
					+ ")");
			return -1;
		}
	}

	/**
	 * Check if an entry for the given concepts exist. The order does not
	 * matter. <br>
	 * <b>Note: If there is an error on database access, this method will return
	 * <code>false</code>.</b>
	 * 
	 * @param c1
	 *            a concept
	 * @param c2
	 *            another concept
	 * @return <code>true</code> if an entry for these concepts exists <b>and</b>
	 *         no database error occures, <code>false</code> otherwise
	 * @see similarity.concepts.matcher.util.Cache#contains(similarity.ontology.Concept,
	 *      similarity.ontology.Concept)
	 */
	public boolean contains(Concept c1, Concept c2) {
		return get(c1, c2) != -1;
	}

	/**
	 * Insert row into table. <br>
	 * <b>Note: If there is an error on database access, this method has no
	 * effect and no exception will be thrown.</b>
	 * 
	 * @param c1
	 *            first concept
	 * @param c2
	 *            second concept
	 * @param sim
	 *            similarity value between 0 and 1 (incl.)
	 * @see similarity.concepts.matcher.util.Cache#put(similarity.ontology.Concept,
	 *      similarity.ontology.Concept, float)
	 */
	public void put(Concept c1, Concept c2, float sim) {
		assert sim >= 0;
		assert sim <= 1;
		assert c1 != null;
		assert c2 != null;
		assert c1.getURI().length() <= MAX_URI_LENGTH;
		assert c2.getURI().length() <= MAX_URI_LENGTH;

		SQLDBAccess.execute("INSERT into " + TABLE + " VALUES ('" + c1.getURI()
				+ "','" + c2.getURI() + "'," + sim + ");");

	}

	/**
	 * Reset similarty table of database. <br>
	 * <b>Note: If there is an error on database access, this method has no
	 * effect and no exception will be thrown.<br>
	 * Caution: All data will be deleted!</b>
	 * 
	 * 
	 * @see similarity.concepts.matcher.util.Cache#reset()
	 */
	public void reset() {

		drop();

		create();

	}

	/**
	 * Drop table.
	 * 
	 * @throws SQLException
	 */
	public void drop() {
		SQLDBAccess.execute("DROP TABLE " + TABLE + ";");
	}

	/**
	 * Create table.
	 * 
	 * @throws SQLException
	 */
	public void create() {
		SQLDBAccess
				.execute("CREATE TABLE " + TABLE + "(" + "c1 VARCHAR("
						+ MAX_URI_LENGTH + ") NOT NULL, " + "c2 VARCHAR("
						+ MAX_URI_LENGTH + ") NOT NULL, "
						+ "sim FLOAT UNSIGNED NOT NULL, "
						+ "PRIMARY KEY(c1,c2)" + ");");
	}
}
