/**
 * This package provides a framework to calculate similarity between concepts.
 */
package similarity.concepts;

import similarity.concepts.matcher.Matcher;
import similarity.concepts.matcher.semantic.*;
import similarity.concepts.matcher.string.StringMatcherFactory;
import similarity.concepts.matcher.util.CombinedMatcher;
import similarity.concepts.matcher.util.MultiMatcher;
import similarity.ontology.Concept;

/**
 * Engine for calculating concept similarities with a given Matcher.
 * This class also provides a default matcher which combines a good selection
 * of ontolgy and string based matchers.
 * 
 * @author Rafael Grote <grote@inf.fu-berlin.de>
 */
public class Engine
{
    /**
     * Default matcher.
     * Selection of several ontolgy based (75%) and string based (25%) matchers.
     */
    public final static Matcher DEFAULT_MATCHER = 
        new CombinedMatcher
        (
                // ontology based
                new MultiMatcher()
                    .addMatcher(new LeacockChodorow(),3)
                    .addMatcher(new WuPalmer(),5)
                    //.addMatcher(new Rada())
                    //.addMatcher(new JiangConrath(),1)
                    .addMatcher(new Lin(),2),
                
                // string based
                new MultiMatcher()
                    .addMatcher(StringMatcherFactory.createLevenshteinMatcher(),2)
                    .addMatcher(StringMatcherFactory.createSoundexMatcher(),1),
                
                // weight of the ontology based matchers
                0.75f
        );
    
    /**
     * Matcher to be used.
     */
    private final Matcher matcher;
    
    /**
     * Matching engine, which uses DEFAULT_MATCHER
     */
    public Engine()
    {
        this(DEFAULT_MATCHER);
    }
    
    /**
     * Engine for calculating concept similarities with the given Matcher.
     * 
     * @param matcher for similarity calculations
     */
    public Engine(Matcher matcher)
    {
        this.matcher = matcher;
    }
    
    /**
     * Calculate similarity of two nodes
     */
    public Result calcSim(Concept c1, Concept c2)
    {
        return new Result(c1,c2,matcher.calcSim(c1,c2));
    }
}
