/**
 * All matchers contained in this package use ontology based measures.
 */
package similarity.concepts.matcher.semantic;



import similarity.concepts.matcher.Matcher;
import similarity.ontology.*;



/**
 * This contains the basic functionality for an ontology based matcher.
 * All matchers in thias package are derived from it.
 * 
 * @author Rafael Grote <grote@inf.fu-berlin.de>
 */
public abstract class AbstractTaxonomicMatcher implements Matcher
{
    /**
     * Searches for a common taxonomy. The similarity is calculated by 
     * calcSemanticSim (implemtented in each matcher).
     * 
     * @see Matcher#calcSim(Concept, Concept)
     */
    public final float calcSim(Concept c1, Concept c2)
    {
        Taxonomy t = Taxonomy.commonTaxonomy(c1,c2);
        
        // taxonomic matchers need a common taxonomy
        if (t == null) return 0;
        
        try
        {
            return (float) normalize(calcSemanticSim(c1,c2,t), maxValue(t));
        }
        // the exceptions should only occur if there is any bug
        catch (ConceptNotFoundException e)
        {
            // should not happen, since the concepts are in the taxonomy!!!
            assert false;
            return 0;
        }
        catch (NoPathException e)
        {
            // should not happen, since the concepts do have a common taxonomy!!!
            assert false;
            return 0;
        }
    }
    
    /**
     * Unnormalized similarity.
     *  
     * Subclasses must implement this method.
     * Subclasses can assume that all parameters are different from null.
     * Further it can be assumed that the concepts really belong to the taxonomy.
     * Subclasses do not need to care about ontology related exceptions.
     * 
     * @param c1 a concept
     * @param c2 another concept2
     * @param t common taxonomy
     * @return similarity between the given concepts
     * @throws ConceptNotFoundException if the concepts do not belong to the taxonomy t 
     * @throws NoPathException if there is no path between the concepts
     */
    protected abstract double calcSemanticSim(Concept c1, Concept c2, Taxonomy t)
        throws ConceptNotFoundException, NoPathException;

    /**
     * Maximum result value for a measure inside the given taxonomy.
     * 
     * Subclasses must implement this.
     * 
     * @param taxonomy
     * @return maximum value of calcSemanticSim(x,y,taxonomy) for any x and y
     */
    protected abstract double maxValue(Taxonomy taxonomy);
    
    /**
     * Simple normalization function.
     * 
     * Subclasses which need a better normalization, should override this method.
     * 
     * @param value value to normalize
     * @param max maximum possible value (must be positive)
     * @return value / max
     */
    protected double normalize(double value, double max)
    {
        assert max > 0; 
        assert value >= 0;
        assert value <= max;
        
        return value / max;
    }

    @Override
    public String toString()
    {
        return this.getClass().getSimpleName();
    }
}
