package similarity.concepts.matcher.util;

import similarity.concepts.matcher.Matcher;
import similarity.ontology.Concept;


/**
 * This is a special matcher which reads similarity values from a cache
 * instead of calculating them on the fly.
 * 
 * Additionally it is able to incrementally fill the cache.
 * 
 * @author Rafael Grote <grote@inf.fu-berlin.de>
 */
public class CachedMatcher implements Matcher
{
    /**
     * Used similarity cache.
     */
    private final Cache cache;
    
    /**
     * Alternative matcher if no value found in cache.
     */
    private final Matcher alternative;
    
    /**
     * Says if the cache should be filled with new values from the
     * alternative matcher if the searched value was net yet existing.
     */
    private final boolean fillcache;
  
    /**
     * Create CachedMatcher without alternative matcher.
     * 
     * @param cache similarity cache
     */
    public CachedMatcher(Cache cache)
    {
        this(cache,null,false);
    }
    
    /**
     * An alternative matcher can be specified. It calculates similarity
     * for concept combinations which are not yet stored in the cache. Set
     * fillcache <code>true</code> if the new values should be saved into cache.
     * <b>Note: You need write access to the underlieing database for that.</b>
     * 
     * If you set the alternative Matcher to <code>null</code>, calcSim() will
     * return 0 for all unkown combinations of concepts. The value of fillcache
     * will be ignored in that case.
     * 
     * @param cache similarity cache
     * @param alternative an alternative matcher
     * @param fillcache <b>true</b> if missing values should be added to the cache,
     *                  <b>false</b> if not
     */
    public CachedMatcher(Cache cache, Matcher alternative, boolean fillcache)
    {
        this.cache = cache;
        this.alternative = alternative;
        this.fillcache = fillcache;
    }

    /**
     * Similarity won't be calculated, but queried from the cache.
     * If no result is found 0 or (if defined) the result of the given
     * alternative matcher is returned.
     * 
     * If an alternative matcher was defined and fillcache is true the newly
     * calculated similarity value will be stored in the cache.
     * 
     * @return cached similarity,
     *         0 if no value found
     *         
     * @see similarity.concepts.matcher.Matcher#calcSim(Concept, Concept)
     */
    public float calcSim(Concept c1, Concept c2)
    {
        float sim = cache.get(c1,c2);
        
        // no value found in cache?
        if (sim < 0)
        {
            // use alternative if existing
            if (alternative != null)
            {
                sim = alternative.calcSim(c1,c2);
                
                // fill cache?
                if (fillcache)
                    cache.put(c1,c2,sim);
            }
            // 0 if no alternative
            else
                sim = 0;                
        }   
        
        return sim;
    }
    
    @Override
    public String toString()
    {
        return "Cached ("+cache.getClass().getSimpleName()+"/"+alternative+")";
    }
}