package similarity.concepts.matcher.util;


import similarity.concepts.matcher.Matcher;
import similarity.ontology.Concept;


/**
 * Combines a selection of semantic and string matchers and weights both groups.
 *  
 * @author Rafael Grote <grote@inf.fu-berlin.de>
 *
 */
public class CombinedMatcher implements Matcher
{
    /**
     * The two parts of the query.
     */
    private final MultiMatcher ontMatcher, strMatcher;
    
    /**
     * The weights for each part.
     */
    private final float ontWeight, strWeight;
    
    /**
     * Create individual combined matcher.
     * The weight of string based matcher is calculated by 1 - ontWeight
     * 
     * @param ontMatcher set of ontology based matchers
     * @param strMatcher set of string based matchers
     * @param ontWeight weight of ontology based matchers, must be a value between 0 and 1
     * 
     * @throws IllegalArgumentException if ontWeight is not between 0 and 1
     */
    public CombinedMatcher(MultiMatcher ontMatcher, MultiMatcher strMatcher, float ontWeight)
    {
        if (ontWeight < 0 || ontWeight > 1)
            throw new IllegalArgumentException("ontWeight must be in range 0..1");
        
        this.ontMatcher = ontMatcher;
        this.strMatcher = strMatcher;
        this.ontWeight = ontWeight;
        this.strWeight = 1f - ontWeight;
    }

    /**
     * Calculate similarity.
     */
    public float calcSim(Concept n1, Concept n2)
    {
        return (ontWeight*ontMatcher.calcSim(n1,n2)) + (strWeight*strMatcher.calcSim(n1,n2));
    }
    
    @Override
    public String toString()
    {
        return "Combined (Semantic("+ontWeight+"): "+ontMatcher+", String("+strWeight+"): "+strMatcher+")";
    }
}
