/**
 * Provides classes and interfaces for calculating similarity between concepts.
 */
package similarity.concepts.matcher.util;

import java.util.ArrayList;
import java.util.Collection;

import similarity.concepts.matcher.Matcher;
import similarity.ontology.Concept;


/**
 * Summarizes multiple weighted matchers. 
 * 
 * @author Rafael Grote <grote@inf.fu-berlin.de>
 *
 */
public class MultiMatcher implements Matcher
{
    /**
     * Matchers to be used (weighted)
     */
    private final Collection<WeightedMatcher> matchers;
    
    /**
     * Create empty MultiMatcher.
     * <b>Warning: At least one matcher must be added,
     *             before calcSim() is called!</b>
     */
    public MultiMatcher()
    {
        this(new ArrayList<WeightedMatcher>());
    }
    
    /**
     * Create MultiMatcher from the given Collection of matchers.
     * 
     * @param matchers initial collection of matchers
     */
    public MultiMatcher(Collection<WeightedMatcher> matchers)
    {
        this.matchers = matchers;
    }

    /**
     * Add a weighted matcher.
     * 
     * @param matcher matcher to be added
     * @param weight factor the matcher will be weighted with in this query, must be positive
     * @throws IllegalArgumentException if weight <= 0
     * @return this to allow cascaded calls
     */
    public MultiMatcher addMatcher(Matcher matcher, float weight)
    {
        matchers.add(new WeightedMatcher(matcher,weight));
        
        return this;
    }

    /**
     * Calculate similarity now.
     * 
     * @return similarity between n1 and n2
     * @throws IllegalStateException if called before a matcher has benn added
     */
    public float calcSim(Concept n1, Concept n2)
    {
        if (matchers.isEmpty())
            throw new IllegalStateException();
        
        float weightedSum = 0f, totalWeight = 0f;
        
        for (WeightedMatcher matcher : matchers)
        {
            weightedSum += matcher.calcSim(n1,n2);
            totalWeight += matcher.getWeight();
        }
        
        return weightedSum / totalWeight;
    }

    @Override
    public String toString()
    {
        return matchers.toString();
    }
}
