/**
 * Provides several special matchers.
 */
package similarity.concepts.matcher.util;


import similarity.concepts.matcher.Matcher;
import similarity.ontology.Concept;

/**
 * A wrapper for matchers, which adds a weight.
 * 
 * @author Rafael Grote <grote@inf.fu-berlin.de>
 *
 */
public class WeightedMatcher implements Matcher
{
    /**
     * weight of the matcher
     */
    private final float weight;

    /**
     * wrapped matcher
     */
    private final Matcher matcher;
    
    /**
     * @param weight factor for the matching
     * @param matcher wrapped matcher
     * @throws IllegalArgumentException if weight <= 0
     */
    public WeightedMatcher(Matcher matcher, float weight)
    {
        if (weight <= 0f)
            throw new IllegalArgumentException("weight must be positive");
        
        this.matcher = matcher;
        this.weight = weight;
    }

    /**
     * Returns the result of the wrapped matcher multiplied with the weight.
     * @return weight * matcher.calcSim(queryNode,resNode)
     * @see similarity.concepts.matcher.Matcher#calcSim(Concept, Concept)
     */
    public float calcSim(Concept c1, Concept c2)
    {
        return weight * matcher.calcSim(c1,c2);
    }
    
    /**
     * @return weight
     */
    public float getWeight()
    {
        return this.weight;
    }
    
    @Override
    public String toString()
    {
        return matcher + "("+weight+")";
    }
}
