package similarity.documents;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import similarity.concepts.Engine;
import similarity.concepts.Result;
import similarity.ontology.Concept;

/**
 * This matcher tries to find for each concept (of both documents) the best
 * matching concept of the other document (inside a section). The pair with
 * highest similarity value is considered best matching. The similarity of
 * a section is the average of the best pairings similarities. The overall
 * similarity is calculated according to the weighting of sections. 
 * 
 * @author Rafael Grote <grote@inf.fu-berlin.de>
 */
public class PairingMatcher<Section extends Enum<Section>> implements Matcher<Section>
{
    /**
     * This map contains <b>all</b> sections as keys and their related
     * weights as value.
     */
    private final Map<Section,Float> weighting;
    
    private final Engine engine;

    /**
     * Create a pairing matcher.
     * A concept matching engine and a set of weights for the sections must
     * be specified.
     * <br>
     * <b>Note: Only sections contained in the weighting map will be
     *          considered for similarity calculation!</b>
     *          
     * @param engine concept matching engine
     * @param weighting <b>all</b> maps and their <b>positve</b> weightings
     */
    public PairingMatcher(Engine engine, Map<Section,Float> weighting)
    {
        this.engine = engine;
        this.weighting = weighting;
    }
    
    /**
     * Calculate overall similarity for given documents.
     * Therefor similarity of all sections is calculated and weighted.
     * 
     * @see similarity.documents.Matcher#similarity(similarity.documents.Document, similarity.documents.Document)
     */
    public similarity.documents.Result<Section> similarity(Document<Section> doc1, Document<Section> doc2)
    {
        float totalWeight = 0;
        float totalSim = 0;

        Map<Section, Float> partialResults = new HashMap<Section, Float>();
        
        // calc partial similarity values
        for (Map.Entry<Section,Float> entry : weighting.entrySet())
        {
            Section section = entry.getKey();
            float weight = entry.getValue();
            
            assert weight > 0;
            
            float sim = similarity(doc1,doc2,section);
            
            // ignore a section, if not present in both documents
          
      
            
            if (sim < 0)
                continue;
            
            totalSim += weight * sim;
            totalWeight += weight;
            
           
            
            partialResults.put(section,sim);
        }
        
        
        
        // calc total similarity
        float sim = (totalWeight==0 ? 0 : totalSim/totalWeight);  
        
        
      
        
        return new similarity.documents.Result<Section>(sim,partialResults);
    }

    /**
     * Calculate similarity for section sec of both documents.
     * 
     * @param doc1 a document
     * @param doc2 another document
     * @param sec a section (present in the given documents)
     * @return similarity value in [0,1],
     *         0 if one section is empty, 
     *         -1 if both sections are empty
     */
    protected float similarity(Document<Section> doc1, Document<Section> doc2, Section sec)
    {
        int count = 0;
        float sum = 0;
        
        Map<Concept,Integer> c1 = doc1.getConcepts(sec);
        Map<Concept,Integer> c2 = doc2.getConcepts(sec);
        
        // Return -1 if both sections empty
        if (c1.isEmpty() && c2.isEmpty())
            return -1;
        
        // If one of the document has no concepts in this section but the
        // other one has, result is 0. (No similarity at all!)
        if (c1.isEmpty() || c2.isEmpty())
            return 0;
        
        // calculate average of bestPairings
        for (Result r : bestPairings(c1,c2,true))
        {
            sum += r.getValue();
            count++;
        }
        
        assert count != 0;
        
        return sum / count;
    }
    
    /**
     * Find best pairing in c2s (highest similarity) for each consept of c1s.
     * 
     * @param c1s a set of concept
     * @param c2s another set of concepts
     * @param both if <code>true</code> the method is invoked a second time with switched arguments
     * @return best pairing and its similarity value
     */
    private Collection<Result> bestPairings(Map<Concept,Integer> c1s, Map<Concept,Integer> c2s, boolean both)
    {
        // return empty result if one map is empty
        if (c1s.isEmpty() || c2s.isEmpty())
            return Collections.emptyList();
        
        Collection<Result> pairings = new ArrayList<Result>(); 
        
        // find best partner for each concept in c1s
        for (Map.Entry<Concept,Integer> c1 : c1s.entrySet())
        {
            Result best = bestPairing(c1.getKey(), c2s);
            
            assert best != null;
            
            pairings.addAll(Collections.nCopies(c1.getValue(),best));
        }
        
        // and now vice versa (if not yet done)
        if (both)
            pairings.addAll(bestPairings(c2s,c1s,false));
        
        return pairings;
    }

    /**
     * Find best pairing in c2s (highest similarity) for c1.
     * 
     * @param c1 a concept
     * @param c2s a set of concepts
     * @return best pairing and its similarity value
     */
    private Result bestPairing(Concept c1, Map<Concept,Integer> c2s)
    {
        // optimization for concepts contained in both docs 
        if (c2s.keySet().contains(c1))
            return new Result(c1,c1,1);
        
        Result best = null;
        
        // get pair with max similarity
        for (Concept c2 : c2s.keySet())
        {
            Result r = engine.calcSim(c1,c2);
            
            if (best==null || best.compareTo(r) < 0)
                best = r;
        }
        
        return best;
    }
}
