/**
 * A representation of an ontology based on Jena.
 */
package similarity.ontology;

import java.util.ArrayList;
import java.util.List;

import com.hp.hpl.jena.ontology.ConversionException;
import com.hp.hpl.jena.ontology.Individual;
import com.hp.hpl.jena.ontology.OntClass;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.util.iterator.ExtendedIterator;
import com.hp.hpl.jena.util.iterator.Filter;
import com.hp.hpl.jena.util.iterator.NullIterator;
import com.hp.hpl.jena.util.iterator.SingletonIterator;
import com.hp.hpl.jena.util.iterator.UniqueExtendedIterator;

import static com.hp.hpl.jena.graph.compose.CompositionBase.countIterator;

/**
 * Encapsulates a Jena OntClass and provides several methods needed by
 * semantic matchers.
 * 
 * <br><b>Note:</b> Some methods need a reasoner to work properly. Thus the
 * the underlying ontology should be loaded as a transitive jena model. 
 * Tested for the <code>OWL_DL_MEM_TRANS_INF</code> ontlogy model
 * specification. 
 * 
 * @author Rafael Grote <grote@inf.fu-berlin.de>
 */
public class Concept
{
    /**
     * Total number of individuals
     */
    private static int INDIVIDUAL_COUNT = -1;
    
    /**
     * Jena representation of this concept.
     */
    private final OntClass ontclass;
    
    /**
     * Cached result of method listTaxonomies().
     * 
     * <b>Warning: May not be initialized. Use listTaxonomies() instead.</b>
     */
    private List<Taxonomy> taxonomies;
    
    /**
     * Create Concept from RDFNode.
     *
     * @param rdfnode a node which represents an OWL concept,
     *                rdfnode.canAs(OntClass) must be true,
     *                must not be null 
     * @throws ConversionException if rdfnode cannot be converted to OntClass
     */
    public Concept(RDFNode rdfnode) throws ConversionException
    {
        assert rdfnode != null;
        assert rdfnode.canAs(OntClass.class);

        this.ontclass = (OntClass) rdfnode.as(OntClass.class);
    }
    
    
    /**
     * Search least common subsumer (lcs) of this and the specified concept.
     * The lcs of two concepts must be a superclass of both. If there are multiple
     * superclasses, the one with the greatest depth (least common) is selected.
     * If one of the concepts is a superclass of the other one, it is the lcs.
     * If the two concepts are equal, they are lcs themself.
     * 
     * @param c other concept
     * @return least common subsumer
     * @throws NoPathException  if there is no path
     */
    public Concept leastCommonSubsumer(Concept c) throws NoPathException
    {
        return searchDistanceAndLCS(c).getConcept();
    }
    
    /**
     * Calculate the distance between this and the specified concept.
     * The distance is defined to be length of shortest path.
     * It will be assumed that the least common subsumer is the root of
     * the shortest path. This measure counts the nodes of the path.
     * Thus the <b>minimum path length is 1</b> (if this.equals(c)). 
     * 
     * @param c other concept
     * @return length of shortest path
     * @throws NoPathException if there is no path
     */
    public int distance(Concept c) throws NoPathException
    {
        return searchDistanceAndLCS(c).getDistance();
    }
    
    /**
     * Get information content of this concept.
     * 
     * The information content of a concept c will be calculated by the
     * formula <b>ic(c) = -log(p(c))</b>, where p(c) is the probability of
     * encountering an instance of c. This includes instances of subconcepts.  
     * 
     * @return value of information content
     */
    public double informationContent()
    {
        // count individuals of this concept
        double freq = listIndividuals(false).size();

        // count individuals in total
        double N = getTotalIndividualCount();

        if (freq == 0 || N == 0)
            return 0;
        
        return -Math.log(freq/N);
    }
    
    /**
     * Count individuals of <b>all</b> concepts. 
     * 
     * @return number of all individuals
     */
    public int getTotalIndividualCount()
    {
        if (INDIVIDUAL_COUNT < 0)
            INDIVIDUAL_COUNT = countIterator(ontclass.getOntModel().listIndividuals());
        
        return INDIVIDUAL_COUNT;
    }
    
    /**
     * Get a list of all taxonomies to which this concept belongs.
     * 
     * @return list of all taxonomies to which this concept belongs
     */
    public List<Taxonomy> listTaxonomies()
    {
        if (taxonomies == null)
        {
            taxonomies = new ArrayList<Taxonomy>();
            ExtendedIterator taxonroots = listRootClasses();
            
            while (taxonroots.hasNext())
                taxonomies.add(Taxonomy.getTaxonomy(new Concept((OntClass) taxonroots.next())));
            
            taxonroots.close();
        }
        
        return taxonomies;
    }
    
    /**
     * List the individuals of this concept.
     * 
     * @param direct if <code>true</code> list <b>only direct</b> individuals of this concept,
     *               if <code>false</code> list <b>all</b> individuals of this concept 
     *                                     (including the individuals of all subconcepts)
     * @return list of subconcepts
     */
    public List<Individual> listIndividuals(boolean direct)
    {
        List<Individual> individuals = new ArrayList<Individual>();
        
        ExtendedIterator it = listInstances(direct);
        
        while (it.hasNext())
            individuals.add((Individual) it.next());
        
        it.close();
        
        return individuals;
    }
    
    /**
     * Get a list of subconcepts.
     *
     * @param direct if <code>true</code> list <b>only direct</b> subconcepts,
     *               if <code>false</code> list <b>all</b> subconcepts 
     *                                     (including the subconcepts of its subconcepts and their ...) 
     * @return list of subconcepts
     */
    public List<Concept> listSubConcepts(boolean direct)
    {
        List<Concept> subconcepts = new ArrayList<Concept>();
        
        ExtendedIterator it = this.ontclass.listSubClasses(direct);
        
        while (it.hasNext())
            subconcepts.add(new Concept((OntClass) it.next()));
        
        it.close();
        
        return subconcepts;
    }
    
    /**
     * Check if the specified concept is a subconcept of this.
     * 
     * @param c other concept
     * @param direct if <code>true</code> only direct subconcepts are considered,
     *               if <code>false</code> all subconcepts are considered
     * @return <code>true</code> if c is a subconcept of this,
     *         <code>false</code> otherwise
     */
    public boolean hasSubConcept(Concept c, boolean direct)
    {
        return ontclass.hasSubClass(c.ontclass,direct);
    }
    
    /**
     * Check if the specified concept is a superconcept of this.
     * 
     * @param c other concept
     * @param direct if <code>true</code> only direct superconcepts are considered,
     *               if <code>false</code> all superconcepts are considered
     * @return <code>true</code> if c is a superconcept of this,
     *         <code>false</code> otherwise
     */
    public boolean hasSuperConcept(Concept c, boolean direct)
    {
        return ontclass.hasSuperClass(c.ontclass,direct);
    }
    
    /**
     * Check if this concept is a leaf.
     * Leafs do not have any subclasses.
     * 
     * @return <code>true</code> if this concept does not have any subclasses,
     *         <code>false</code> otherwise
     */
    public boolean isLeaf()
    {
        // Note that when using a reasoner, all OWL classes have owl:Nothing as a sub-class.
        // So we cant use hasSubClass(). But listSubClasses() will be empty, if there are no
        // subclasses except owl:Nothing.
        
        return isEmpty(ontclass.listSubClasses(false));
    }
    
    /**
     * Get the label of this concept.
     * Returns the local name of the corresponding ontclass.
     * For anonymous concepts the id will be returned. 
     * 
     * @return label
     */
    public String getLabel()
    {
        if (ontclass.isAnon())
            return ontclass.getId().toString();
        else
            return ontclass.getLocalName();
    }
    
    /**
     * Get the URI of this concept.
     * For anonymous concepts this will return <code>null</code>.
     * 
     * @return URI
     */
    public String getURI()
    {
        return ontclass.getURI();
    }
    
    /**
     * Get a unique id for this concept.
     * For concept which have an URI, the URI will be returned. For anonymous
     * concepts the id of the related ontclass will be returned.
     * 
     * @return unique id
     */
    public String getID()
    {
        if (ontclass.isAnon())
            return ontclass.getId().toString();
        else
            return getURI();   
    }
    
       
    /*
     * HELPER
     */
    
    /**
     * Calculate distance to the specified concept, which must be superconcept or equal to this.
     * 
     * @param c superconcept of this
     * @return distance
     * @throws ConceptNotFoundException if concept is not a superclass or equal to this
     */
    protected int distanceToSuperConcept(Concept c) throws ConceptNotFoundException
    {
        SearchResult result = searchSuperConcept(c.new IdentityFilter());
        
        if (result.isEmpty())
            throw new ConceptNotFoundException(c + " is not a superconcept of " + this);
        
        return result.getDistance();
    }
    
    /**
     * Search <b>upwards</b> (in direction of root).
     * Finds the nearest superconcept of this, which is accepted by the specified filter.
     * If this is accepted by the filter, this will be returned. 
     * 
     * @param filter condition for result
     * @return result of search(filter,true)
     * @see Concept#search(Filter, boolean)
     */
    protected SearchResult searchSuperConcept(Filter filter)
    {
        return search(filter,true);
    }

    /**
     * Search <b>downwards</b> (in direction of leaves).
     * Finds the nearest subconcept of this, which is accepted by the specified filter. 
     * If this is accepted by the filter, this will be returned.
     * 
     * @param filter condition for result
     * @return result of search(filter,false)
     * @see Concept#search(Filter, boolean)
     */
    protected SearchResult searchSubConcept(Filter filter)
    {
        return search(filter,false);
    }
    
    /**
     * Search the nearest concepts (including this) accepted by the specified filter.
     * Superclasses (if upward is true) or subclasses (if upward is false) are searched.
     * 
     * @param filter accepts result
     * @param upward if <b>true</b> search in superclasses, if <b>false</b> search in subclasses
     * @return found concept and its distance to this concept, EmptyResult if none found
     */
    private SearchResult search(Filter filter, boolean upward)
    {
        return search(filter,upward,true).get(0);
    }
    
    /**
     * Search concepts (including this) accepted by the specified filter.
     * Superclasses (if upward is true) or subclasses (if upward is false) are searched.
     * If oneresult is set to true the first (= nearest) result or (if no) an empty result will be returned.
     * 
     * @param filter accepts result
     * @param upward if <b>true</b> search in superclasses, if <b>false</b> search in subclasses
     * @param oneresult if <b>true</b> return only the nearest result, if <b>false</b> return all results 
     * @return list of found concepts and their distances to this concept
     */
    private List<SearchResult> search(Filter filter, boolean upward, boolean oneresult)
    {
        ExtendedIterator currentlevel, nextlevel;
        int distance = 0;
        List<SearchResult> results = new ArrayList<SearchResult>();
        
        // start with this concept
        currentlevel = new SingletonIterator(this.ontclass);
        
        while (currentlevel.hasNext())
        {   
            ++distance; 
            nextlevel = NullIterator.instance; // empty

            // find all direct superclasses (subclasses if !upward) 
            while (currentlevel.hasNext())
            {
                OntClass c = (OntClass) currentlevel.next();
                
                // check if we search c
                if (filter.accept(c))
                {
                    results.add(new SearchResult(new Concept(c), distance));
                    
                    if (oneresult) return results;
                }
                
                // get successors 
                ExtendedIterator successors;
                if (upward)
                    successors = c.listSuperClasses(true);
                else
                    successors = c.listSubClasses(true);
                
                // add successors to next level
                nextlevel = nextlevel.andThen(successors);
            }
            
            // close current level
            currentlevel.close();
            // and go into next level
            currentlevel = new UniqueExtendedIterator(nextlevel);
        }
        
        // if nothing was found and we want one result add empty result
        if (oneresult && results.isEmpty()) results.add(new EmptySearchResult(distance));
        
        return results;
    }
    
    /**
     * Special search for distance and lcs. It is good idea to combine this,
     * because lcs is always the rood node of the shortest path. Thus the
     * methods to calculate lcs/distance need the result of each other. 
     * 
     * @param c other concept
     * @return least common subsumer and length of shortest path
     * @throws NoPathException if there is no path
     */
    private SearchResult searchDistanceAndLCS(Concept c) throws NoPathException
    {
        try
        {
            // check if the concepts are equal
            if (this.equals(c)) return new SearchResult(this,1);
            
            // check if one of the concepts is superconcept of the other
            if (this.ontclass.hasSuperClass(c.ontclass,false))
                return new SearchResult(c,this.distanceToSuperConcept(c));
            if (c.ontclass.hasSuperClass(this.ontclass,false))
                return new SearchResult(this,c.distanceToSuperConcept(this));
            
            // find all subsumers (and their distances to this)
            List<SearchResult> subsumers = search(new SuperClassFilter(false).and(c.new SuperClassFilter(false)),true,false);
            
            SearchResult best = new EmptySearchResult(0);
            
            // now calculate the totel distance and select the best result
            for (SearchResult s : subsumers)
            {
                int distance = s.getDistance() + c.distanceToSuperConcept(s.getConcept()) -1;
                
                if (distance < best.getDistance())
                    best = new SearchResult(s.getConcept(),distance);
            }
            
            // no path found?
            if (best.isEmpty()) throw new NoPathException("There is no path between " + this + " and " + c);
            
            return best;
        }
        catch (ConceptNotFoundException e)
        {
            // should not happen
            assert false;
            throw new NoPathException("INTERNAL ERROR");
        }
    }
    
    /**
     * List all roots of the taxonomies, to which this concept belongs.
     * Anonymous root concepts will not be considered.
     */
    protected ExtendedIterator listRootClasses()
    {
        return ontclass.listSuperClasses(false).andThen(new SingletonIterator(ontclass))
                    .filterKeep(new RootFilter().and(new NotAnonymousFilter()));
    }
    
    /**
     * List the instances of this concept class.
     * 
     * @param direct if <code>true</code> list <b>only direct</b> instances of this concept,
     *               if <code>false</code> list <b>all</b> instances of this concept 
     *                                     (including the instances of subconcepts)
     */
    protected ExtendedIterator listInstances(boolean direct)
    {
        ExtendedIterator instances = ontclass.listInstances();
        
        if (!direct) // add instances of all subclasses 
        {
            ExtendedIterator subclasses = ontclass.listSubClasses(false);
            
            while (subclasses.hasNext())
                instances = instances.andThen(((OntClass)subclasses.next()).listInstances());
            
            subclasses.close();
        }
        
        return instances;
    }
    
    /**    
     * This method is only for testing.
     * 
     * @return the Jena OntClass of this Concept
     */
    public OntClass getOntClass()
    {
        return this.ontclass;
    }
    
    
    /**
     * Check if the given iterator is empty.
     * As a side effect the iterator will be closed. 
     * 
     * @param iterator
     * @return !iterator.hasNext()
     */
    public static boolean isEmpty(ExtendedIterator iterator)
    {
        boolean result = !iterator.hasNext();
        iterator.close();
        return result;
    }
    
    
    /*
     * Overridden methods of Object
     */
    
    @Override
    public String toString()
    {
        return getLabel();
    }
    
    @Override
    public int hashCode()
    {
        return getID().hashCode();
    }
    
    @Override
    public boolean equals(Object o)
    {
        if (o instanceof Concept) return equals((Concept) o);
        else return false;
    }
    
    public boolean equals(Concept c)
    {
        return getID().equalsIgnoreCase(c.getID());
    }
    
    
    /*
     * FILTER
     */
    
    /**
     * Accepts concepts, which are subclass of this concept.
     */
    class SubClassFilter extends Filter
    {
        final boolean direct; 
        
        SubClassFilter(boolean direct)
        {
            this.direct = direct;
        }
        
        @Override
        public boolean accept(Object o)
        {
            return Concept.this.ontclass.hasSubClass((OntClass) o, direct);
        }
        
    }
    
    /**
     * Accepts concepts, which are superclass of this concept.
     */
    class SuperClassFilter extends Filter
    {
        final boolean direct; 
        
        SuperClassFilter(boolean direct)
        {
            this.direct = direct;
        }
        @Override
        public boolean accept(Object o)
        {
            return Concept.this.ontclass.hasSuperClass((OntClass) o, direct);
        }
        
    }
    
    /**
     * Accepts only this concept.
     */
    class IdentityFilter extends Filter
    {
        @Override
        public boolean accept(Object o)
        {
            return Concept.this.ontclass.equals(o);
        }
        
    }
    
    /**
     * Accepts concepts, which are root.
     */
    class RootFilter extends Filter
    {
        @Override
        public boolean accept(Object o)
        {
            OntClass oc = (OntClass) o;
            
            return oc.isHierarchyRoot() || !hasNonAnonymousSuperClasses(oc);
        }
        
        private boolean hasNonAnonymousSuperClasses(OntClass oc)
        {
            return !isEmpty(oc.listSuperClasses(false).filterKeep(new NotAnonymousFilter()));
        }
    }

    /**
     * Accepts concepts, which are NOT anonymous.
     */
    class NotAnonymousFilter extends Filter
    {
        @Override
        public boolean accept(Object o)
        {
            return !((OntClass) o).isAnon();
        }
    }
    
    /**
     * Accepts nothing.
     */
    class AlwaysFalseFilter extends Filter
    {
        @Override
        public boolean accept(Object o)
        {
            return false;
        }
    }
}
