package similarity.ontology;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


/**
 * A taxonomy is represented by its root concept.
 * All subconcepts of the root are members of the taxonomy.
 * 
 * @author Rafael Grote <grote@inf.fu-berlin.de>
 */
public class Taxonomy
{
    /**
     * Keep a copy of each instance mapped to its rootconcept.
     * So we can avoid calculating data twice. 
     */
    private static final Map<Concept,Taxonomy> instances = new HashMap<Concept,Taxonomy>();
    
    /**
     * Rootconcept of this taxonomy
     */
    private final Concept root;
    
    /**
     * Depth of the deepest concept of this taxonomy (= number of levels)
     */
    private final int maxDepth;
    
    /**
     * Number of members including root.
     */
    private final int size;

    
    /**
     * Factory for taxonomies. There is only a single Taxonomy-instance for
     * each rootconcept. So we can avoid calculating member data twice.
     * 
     * @param root rootconcept
     * @return taxonomy for the given rootnode.
     */
    public static Taxonomy getTaxonomy(Concept root)
    {
        // create first if not existing
        if (!instances.containsKey(root))
            instances.put(root, new Taxonomy(root));

        return instances.get(root);
    }
    
    
    /**
     * Private constructor.
     * Use getTaxonomy instead.
     * 
     * @param root rootconcept of the taxonomy
     */
    private Taxonomy(Concept root)
    {
        this.root = root;

        List<Concept> subconcepts = root.listSubConcepts(false); 

        this.size = subconcepts.size() + 1;
        
        int max = 1; // depth of root is 1
        
        for (Concept concept : subconcepts)
        {
            if (concept.isLeaf())
            {
                try
                {
                    max = Math.max(depthOf(concept),max);
                }
                catch (ConceptNotFoundException e)
                {
                    assert false; // should not happen!
                }
            }
        }
        
        this.maxDepth = max;
    }
    
    
    /**
     * Check if the specified concept is a member of this taxonomy.
     * 
     * @param c a concept
     * @return <b>true</b> if c is a member of this taxonomy,
     *         <b>false</b> otherwise
     */
    public boolean contains(Concept c)
    {
        return c.hasSuperConcept(root,false);
    }
    
    /**
     * Get depth of the specified concept in this ontology.
     * The depth of a concept is the distance to the root node. Root nodes have a depth value of 1.
     * If there are multiple possible routes to the root node, the shortest route will be taken. 
     * 
     * @return depth of the specified concept in this taxonomy
     * @throws ConceptNotFoundException if c is not a member of this taxonomy 
     */
    public int depthOf(Concept c) throws ConceptNotFoundException
    {
        return c.distanceToSuperConcept(root); 
    }
    
    /**
     * Get the maximum depth of this taxonomy.
     * This value is equal to the number of levels in this taxonomy.
     * 
     * @return depth of the deepest concept
     */
    public int getMaxDepth()
    {
        return maxDepth;
    }
    
    /**
     * Count members.
     * 
     * @return number of members
     */
    public int getSize()
    {
        return size;
    }
    
    /**
     * Get root of this taxonomy.
     * 
     * @return rootconcept
     */
    public Concept getRoot()
    {
        return root;
    }
    
    @Override
    public String toString()
    {
        return "[" + root + "]";
    }
    
    @Override
    public boolean equals(Object o)
    {
        if (o instanceof Taxonomy) return equals((Taxonomy) o);
        else return false;
    }
    
    /**
     * One taxonomy is equal to another, if both have the same rootconcept.
     * 
     * @param t another taxonomy
     * @return result of root.equals(t.root)
     */
    public boolean equals(Taxonomy t)
    {
        return root.equals(t.root);
    }
    
    @Override
    public int hashCode()
    {
        return root.hashCode();
    }

    /**
     * Find a common taxonomy for c1 and c2.
     * If there is no, null will be returned. If there are multiple results,
     * the taxonmy of the lcs is used. If the lcs has multiple taxonomies too,
     * the one with the greatest maximum depth is selected
     * 
     * @param c1 concept1
     * @param c2 concept2
     * @return one of the common taxonomies or null if there is none
     */
    public static Taxonomy commonTaxonomy(Concept c1, Concept c2)
    {
        List<Taxonomy> common = listCommonTaxonomies(c1,c2);
        
        if (common.isEmpty())
            return null;
        
        if (common.size() == 1)
            return common.get(0);
        
        try
        {
            common = c1.leastCommonSubsumer(c2).listTaxonomies();
        }
        catch (NoPathException e)
        {
            // should not happen if there is a common taxonomy
            assert false;
        }
        
        assert common.size() > 0;
        
        int maxDepth = Integer.MIN_VALUE;
        Taxonomy deepest = null;
        
        for (Taxonomy taxonomy : common)
        {
            if (taxonomy.getMaxDepth() > maxDepth)
            {
                maxDepth = taxonomy.getMaxDepth();
                deepest = taxonomy;
            }
        }
        
        assert deepest != null;
        
        return deepest;
    }


    /**
     * Find common taxonomies of c1 and c2.
     * 
     * @param c1 concept1
     * @param c2 concept2
     * @return all taxonmies in which both concepts are members
     */
    public static List<Taxonomy> listCommonTaxonomies(Concept c1, Concept c2)
    {
        List<Taxonomy> common = new ArrayList<Taxonomy>();
        
        for (Taxonomy t1 : c1.listTaxonomies())
            if (c2.listTaxonomies().contains(t1)) common.add(t1);
        
        return common;
    }

}
