
package tools;

import java.util.*;
import java.util.concurrent.Callable;
import java.util.concurrent.locks.ReentrantReadWriteLock;


public class ConcurrentCache<KEY, VALUE>
{
	/* | = - = - = - = - = - = - = - = - = - = /-||=||-\ = - = - = - = - = - = - = - = - = - = | */
	/* | = == === ==== ===== ====== ======= === Fields ==== ======= ====== ===== ==== === == = | */
	/* | = - = - = - = - = - = - = - = - = - = \-||=||-/ = - = - = - = - = - = - = - = - = - = | */

	private final Map<KEY, VALUE>					innerCache	= new Hashtable<KEY, VALUE>();
	private final ReentrantReadWriteLock			globalLock	= new ReentrantReadWriteLock();
	private final Map<KEY, ReentrantReadWriteLock>	entryLocks	= new Hashtable<KEY, ReentrantReadWriteLock>();

	/* | = - = - = - = - = - = - = - = - = - = /-||=||-\ = - = - = - = - = - = - = - = - = - = | */
	/* | = == === ==== ===== ====== ======= === Methods === ======= ====== ===== ==== === == = | */
	/* | = - = - = - = - = - = - = - = - = - = \-||=||-/ = - = - = - = - = - = - = - = - = - = | */

	public VALUE get(KEY key) {
		// Preconditions:
		assert key != null;

		// Implementation:
		if (isWriteLocked(key))
			return innerGet(key);

		readLock(key);
		try {
			return innerGet(key);
		}
		finally {
			readUnLock(key);
		}
	}

	private VALUE innerGet(KEY key) {
		// Preconditions:
		assert key != null;

		// Implementation:
		return innerCache.get(key);
	}

	public Iterable<VALUE> entries() {
		// Implementation:
		if (globalLock.isWriteLockedByCurrentThread())
			return innerValues();

		globalLock.readLock().lock();
		try {
			return innerValues();
		}
		finally {
			globalLock.readLock().unlock();
		}
	}

	private Collection<VALUE> innerValues() {
		return Collections.unmodifiableCollection(innerCache.values());
	}

	private ReentrantReadWriteLock getLock(KEY key) {
		// Preconditions:
		assert key != null;

		// Implementation:�

		// Erstmal davon aufgehen, da� das Lock schon exitiert:
		globalLock.readLock().lock();
		try {
			final ReentrantReadWriteLock readWriteLock = entryLocks.get(key);
			if (readWriteLock != null)
				return readWriteLock;
		}
		finally {
			globalLock.readLock().unlock();
		}

		// Falls nicht, das Lock anlegen:
		globalLock.writeLock().lock();
		try {
			ReentrantReadWriteLock result = entryLocks.get(key);
			if (!entryLocks.containsKey(key))
				entryLocks.put(key, result = new ReentrantReadWriteLock());

			return result;
		}
		finally {
			globalLock.writeLock().unlock();
		}
	}

	private boolean isWriteLocked(KEY key) {
		return getLock(key).isWriteLockedByCurrentThread();
	}

	// ---------------------------------------------------------------------------------------------

	public void remove(KEY key) {
		// Preconditions:
		assert key != null;

		// Implementation:
		if (isWriteLocked(key))
			innerRemove(key);
		else {
			writeLock(key);
			try {
				innerRemove(key);
			}
			finally {
				writeUnLock(key);
			}
		}
	}

	private void innerRemove(KEY key) {
		// Preconditions:
		assert key != null;

		// Implementation:
		innerCache.remove(key);
	}

	public void put(KEY key, VALUE value) {
		// Preconditions:
		assert key != null;
		assert value != null;

		// Implementation:
		if (isWriteLocked(key))
			innerPut(key, value);
		else {
			writeLock(key);
			try {
				innerPut(key, value);
			}
			finally {
				writeUnLock(key);
			}
		}
	}

	private void innerPut(KEY key, VALUE value) {
		// Preconditions:
		assert key != null;
		assert value != null;

		// Implementation:
		innerCache.put(key, value);
	}

	public void forEach(ForEach<VALUE> forEach) {
		// Preconditions:
		assert forEach != null;

		// Implementation:
		if (globalLock.isWriteLockedByCurrentThread())
			innerForEach(forEach);
		else {
			globalLock.writeLock().lock();
			try {
				innerForEach(forEach);
			}
			finally {
				globalLock.writeLock().unlock();
			}
		}
	}

	private void innerForEach(ForEach<VALUE> forEach) {
		// Preconditions:
		assert forEach != null;

		// Implementation:
		for (VALUE value : innerCache.values())
			forEach.each(value);
	}

	public VALUE getOrCreate(KEY key, Callable<VALUE> producer) {
		// Preconditions:
		assert key != null;
		assert producer != null;

		// Implementation:

		// Existiert key bereits?
		VALUE value = get(key);
		if (value != null)
			return value;

		// Neu anlegen
		if (isWriteLocked(key))
			return innerGetOrCreate(key, producer);

		writeLock(key);
		try {
			return innerGetOrCreate(key, producer);
		}
		finally {
			writeUnLock(key);
		}
	}

	private VALUE innerGetOrCreate(KEY key, Callable<VALUE> producer) {
		// Preconditions:
		assert key != null;
		assert producer != null;

		// Implementation:
		VALUE result = innerGet(key);
		if (result == null) {
			try {
				innerPut(key, result = producer.call());
			}
			catch (Exception e) {
				throw new RuntimeException(e);
			}
		}

		return result;
	}

	// ---------------------------------------------------------------------------------------------

	private void readLock(KEY key) {
		getLock(key).readLock().lock();
	}

	private void readUnLock(KEY key) {
		getLock(key).readLock().unlock();
	}

	public void writeLock(KEY key) {
		getLock(key).writeLock().lock();
	}

	public void writeUnLock(KEY key) {
		getLock(key).writeLock().unlock();
	}
}
